/*
    $Id: $

    Copyright 2002 - 2005

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    cpuspeed.cc

    Author:
        Carl Thompson <cet@carlthompson.net>

    Contributors:
        Enrico Tassi <tassi@cs.unibo.it>

    This program is only for computers with Linux kernels compiled with
    CPUFreq.  You must have a CPU that supports frequency and/or voltage
    scaling via CPUFreq to use this program.  Your kernel must be compiled to
    support the "userspace" CPUFreq governor and the "sysfs" interface on
    Linux 2.6 or the "proc" interface on Linux 2.4.

    I use this program on my eMachines M5305 and Acer Ferrari 3200 laptops to
    increase battery life and control performance.
*/

const char VERSION[] = "1.2.1";

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/errno.h>
#include <stdarg.h>
#include <time.h>
#include <sys/utsname.h>
#include <libgen.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/sysinfo.h>

// Maximum number of speed steps. Must be 2 or more. You don't need to change
// this. Even if your CPU supports more speed steps the program will use up to
// this number of them evenly spaced.
const unsigned MAX_SPEEDS  = 15;

// Minimum speed step supported in KHz
const unsigned MIN_STEP = 25000;

const char SYSFS_MIN_SPEED_FILE[] =
    "/sys/devices/system/cpu/cpu%u/cpufreq/scaling_min_freq";

const char SYSFS_MAX_SPEED_FILE[] =
     "/sys/devices/system/cpu/cpu%u/cpufreq/scaling_max_freq";

const char SYSFS_CURRENT_SPEED_FILE[] =
     "/sys/devices/system/cpu/cpu%u/cpufreq/scaling_setspeed";

const char SYSFS_GOVERNOR_FILE[] =
     "/sys/devices/system/cpu/cpu%u/cpufreq/scaling_governor";

const char SYSFS_USERSPACE[] =
    "userspace";

const char * PROC_STAT_FILE =
    "/proc/stat";

#ifdef WITH_PROC
const char PROC_MIN_SPEED_FILE[] =
    "/proc/sys/cpu/%u/speed-min";

const char PROC_MAX_SPEED_FILE[] =
    "/proc/sys/cpu/%u/speed-max";

const char PROC_CURRENT_SPEED_FILE[] =
    "/proc/sys/cpu/%u/speed";

const char PROC_GOVERNOR_FILE[] =
    "/proc/cpufreq";

const char PROC_USERSPACE[] =
    "0%0%100%userspace";
#endif

#ifdef DEBUG
const char * FAKE_STAT_FILE =
    "/tmp/cpuspeed/stat";

const char FAKE_MIN_SPEED_FILE[] =
    "/tmp/cpuspeed/%u/speed-min";

const char FAKE_MAX_SPEED_FILE[] =
    "/tmp/cpuspeed/%u/speed-max";

const char FAKE_CURRENT_SPEED_FILE[] =
    "/tmp/cpuspeed/%u/speed";

const char FAKE_GOVERNOR_FILE[] =
    "/tmp/cpuspeed/%u/governor";

const char FAKE_USERSPACE[] =
    "userspace";
#endif

const char * MIN_SPEED_FILE;
const char * MAX_SPEED_FILE;
const char * CURRENT_SPEED_FILE;
const char * GOVERNOR_FILE;
const char * USERSPACE;
const char * STAT_FILE;

// defines what info we care about for each speed step
struct
{
    unsigned khz;
    //unsigned volts;
    //unsigned fsb;
} speeds[MAX_SPEEDS + 1];
unsigned current_speed; // current speed step
unsigned last_step; // lowest speed step

// whether to account nice time as user time (true) or idle time (false)
bool include_nice = false;

// which CPU are we controlling
unsigned cpu = 0;

// display an error message and exit the program
void
die(bool system_error, const char *fmt, ...)
{
    fprintf(stderr, "Error: ");

    va_list ap;
    // get variable argument list passed
    va_start(ap, fmt);
    // display message passed on stderr
    vfprintf(stderr, fmt, ap);
    va_end(ap);

    fprintf(stderr, "\n");
    if (system_error)
        fprintf(stderr, "Error: %s\n", strerror(errno));

    exit(1);
}

// read a line from a file
void
read_line(const char * filename, char * line, unsigned len)
{
    FILE * fp = fopen(filename, "r");
    if (!fp)
        die(true, "Could not open file for reading: %s", filename);
    if ( (!fgets(line, len, fp)) )
        die(true, "Could not read from file: %s", filename);
    fclose(fp);
}

// write a line to a file
void
write_line(const char * filename, const char *fmt, ...)
{
    FILE * fp = fopen(filename, "w+");
    if (!fp)
        die(true, "Could not open file for writing: %s", filename);
    va_list ap;
    // get variable argument list passed
    va_start(ap, fmt);
    if (vfprintf(fp, fmt, ap) < 0)
        die(true, "Could not write to file: %s", filename);
    va_end(ap);
    fclose(fp);
}

// read an integer value from a file
unsigned
read_value(const char * filename)
{
    char line[256];
    read_line(filename, line, sizeof line);
    return atoi(line);
}

// get the current CPU speed
unsigned get_speed()
{
    return read_value(CURRENT_SPEED_FILE);
}

// set the current CPU speed
void set_speed(unsigned value)
{
#ifdef DEBUG
    fprintf(stderr, "[cpu%u] Setting speed to: %uKHz\n", cpu, value);
#endif
    write_line(CURRENT_SPEED_FILE, "%u\n", value);
    // give CPU / chipset voltage time to settle down
    usleep(10000);
}

// This code smoothly transitions the CPU speed from 'current' to 'target'
// instead of jumping directly to the new speed because some AMD Mobile
// Athlon parts seem to choke on large differentials causing kernel panics.
void set_speed(unsigned current, unsigned target)
{
    if (current == target)
        return;
    int delta = (current > target) ? -1 : 1;
    do
    {
        current += delta;
        set_speed(speeds[current].khz);
    }
    while (current != target);
}


inline unsigned MIN(unsigned a, unsigned b) { return a > b ? b : a; }
inline unsigned MAX(unsigned a, unsigned b) { return a > b ? a : b; }

// the minimum and maximum speed that we are allowed to set
unsigned min_speed = 0, max_speed = UINT_MAX;

// get the speed steps supported by the CPU
void
get_supported_speeds()
{
    unsigned min = read_value(MIN_SPEED_FILE);
    unsigned max = read_value(MAX_SPEED_FILE);

#ifdef DEBUG
    fprintf(stderr, "[cpu%u] Mimimum speed supported: %uKHz\n", cpu, min);
    fprintf(stderr, "[cpu%u] Maximum speed supported: %uKHz\n", cpu, max);
#endif
    min = MAX(min, min_speed);
    max = MIN(max, max_speed);

    unsigned step, real, retries;
    const unsigned MAX_STEP = MIN_STEP << 8;
    for (step = MIN_STEP; step <= MAX_STEP; step *= 2)
    {
        min = MAX(min, step);
        if (max <= min)
            die(false, "No speed steps could be determined!");

        // Go to max speed if we are not already there
        for (unsigned current = get_speed(); current <= max; current += step)
                set_speed(current);
        set_speed(max);

        // this code is a hack to get the various speed steps supported by the
        // CPU by looping from the maximum speed to the minimum speed and trying
        // to set every possible speed divisible by step!
        speeds[0].khz = max;
        current_speed = 0;
        for (unsigned current = max - step; current > min - step; current -= step)
        {
		retries = 0;
                current = MAX(current, min);
		while (retries < 3) {
			set_speed(current);
			real = get_speed();
			if (real == current || real == speeds[current_speed].khz)
				break;
			retries++;
		}
                if (real != speeds[current_speed].khz)
                {
                speeds[++current_speed].khz = real;
                if (current_speed + 1 == MAX_SPEEDS)
                        break;
                }
        }
        if (current_speed + 1 != MAX_SPEEDS)
                break;
    }
    if (step > MAX_STEP)
        die(false, "Detected more speed steps than this program can handle?!");

    speeds[current_speed + 1].khz = 0;

    // the last step is the lowest found speed
    last_step = current_speed;

#ifdef DEBUG
    fprintf(stderr, "[cpu%u] Available speeds:\n", cpu);
    for (unsigned speed = 0; speeds[speed].khz; speed++)
        fprintf(stderr, "[cpu%u]  %2u: %9uKHz\n", cpu, speed, speeds[speed].khz);
#endif
}

// gets the elapsed total time and elapsed idle time since it was last called
void
get_times(unsigned long & total_elapsed, unsigned long & idle_elapsed)
{
    FILE * fp = fopen(STAT_FILE, "r");
    if (!fp)
        die(true, "Could not open %s for reading!", STAT_FILE);

    static char search[8];
    static size_t searchlen = 0;
    if (!searchlen)
    {
        snprintf(search, sizeof search, "cpu%u ", cpu);
        searchlen = strlen(search);
#ifdef DEBUG
        fprintf(stderr, "[cpu%u] Looking for CPU line starting with: \"%s\"\n", cpu, search);
#endif
    }
    bool found;
    char line[256];
    while ( (found = fgets(line, sizeof line, fp)) && strncmp(line, search, searchlen) )
    ;

    fclose(fp);

    if (!found)
    {
#ifdef DEBUG
        fprintf(stderr, "[cpu%u] Could not find \'%s\' line in %s.\n", cpu, search, STAT_FILE);
#endif
        return;
    }

    char what[32];
    unsigned long user_time, nice_time, system_time, idle_time, wait_time = 0;
    sscanf(
        line, "%s %lu %lu %lu %lu %lu", what, &user_time, &nice_time,
        &system_time, &idle_time, &wait_time
    );

    if (!include_nice)
	 idle_time += nice_time;
    else
	 user_time += nice_time;

    // count IO wait time as idle time
    idle_time += wait_time;

    unsigned long total_time = user_time + system_time + idle_time;
    static unsigned long last_total_time = 0, last_idle_time = 0;
    total_elapsed = total_time - last_total_time;
    last_total_time = total_time;
    idle_elapsed = idle_time - last_idle_time;
    last_idle_time = idle_time;

#ifdef DEBUG
    fprintf(stderr, "[cpu%u] time: %lu    idle: %lu\n", cpu, total_elapsed, idle_elapsed);
#endif
}

// resets the elapsed total time and elapsed idle time counters
void reset_times()
{
    unsigned long dummy1, dummy2;
    get_times(dummy1, dummy2);
}

// are we currently dynamically scaling the CPU or at min or max?
enum Mode { SPEED_DYNAMIC, SPEED_MIN, SPEED_MAX } mode;

// if CPU idle percentage is below this, CPU will be set to fastest speed
unsigned clock_up_idle_fast= 10;

// if CPU idle percentage is below this, CPU will be set to next higher speed
// if CPU idle percentage is above this, CPU will be set to next lower speed
unsigned idle_threshold = 25;

// if this is set (via command line) throttle down CPU to minimum if
// temperature is too high
const char * temperature_filename = 0;
unsigned max_temperature;

// if this is set (via command line) throttle down CPU to minimum if
// AC power is disconnected
const char * ac_filename = 0;

// if this is true then maximize speed when AC connected
bool max_speed_on_ac = false;

// if this is true then minimize speed when AC disconnected
bool min_speed_on_battery = true;

bool check_cpu = false, check_therm = false, check_ac = false;

// handles the periodic check of idle and setting CPU speed
void
alarm_handler(int)
{
    unsigned old_speed = current_speed;

    // current state
    Mode state = mode;

    static bool on_ac = true;

    char line[256], *p;

    if (mode == SPEED_DYNAMIC)
    {
        // check to see if AC power is disconnected
        if (ac_filename && check_ac)
        {
            read_line(ac_filename, line, sizeof line);
            if (strstr(line, "off-line"))
            {
                on_ac = false;
#ifdef DEBUG
                fprintf(stderr, "[cpu%u] AC is off-line\n", cpu);
#endif
            }
            else
            {
                on_ac = true;
#ifdef DEBUG
                fprintf(stderr, "[cpu%u] AC is on-line\n", cpu);
#endif
            }
        }

        if (max_speed_on_ac && on_ac)
            state = SPEED_MAX;
        else if (!on_ac && min_speed_on_battery)
            state = SPEED_MIN;

        // check that we are not getting too hot
        if (temperature_filename && check_therm && state != SPEED_MIN)
        {
            read_line(temperature_filename, line, sizeof line);
            p = strpbrk(line, "0123456789");
            if (!p)
                die(false, "Could not find temperature in file: %s", temperature_filename);
#ifdef DEBUG
        fprintf(stderr, "[cpu%u] temp: %ld\n", cpu, strtol(p, 0, 10));
#endif
            if (strtol(p, 0, 10) > (long)max_temperature)
                state = SPEED_MIN;
        }
    }

    // figure out what our current speed should be
    switch(state)
    {
        case SPEED_DYNAMIC:
        {
            // if it's not yet time to check CPU then don't
            if (!check_cpu)
                break;

            unsigned long elapsed_time, idle_time;
            unsigned idle_percent;

            // get the elapsed and idle times since we last checked
            get_times(elapsed_time, idle_time);

            if (elapsed_time > 0)
            {
                idle_percent = idle_time * 100 / elapsed_time;

                if (idle_percent <= clock_up_idle_fast)
                    current_speed = 0;
                else if (idle_percent < idle_threshold && current_speed > 0)
                    current_speed--;
                else if (idle_percent > idle_threshold && speeds[current_speed + 1].khz != 0)
                    current_speed++;
#ifdef DEBUG
                fprintf(stderr, "[cpu%u] Idle percent: %.2u\n", cpu, idle_percent);
#endif
            }
            break;
        }

        case SPEED_MIN:
            current_speed = last_step;
            break;

        case SPEED_MAX:
            current_speed = 0;
            break;
    }

    // if the last set speed is not what it currently should be, set it
    if (current_speed != old_speed)
    {
#ifdef DEBUG
        fprintf(
            stderr, "[cpu%u] old speed: %uKHz     new speed: %uKHz\n",
            cpu, speeds[old_speed].khz, speeds[current_speed].khz
        );
#endif
        set_speed(old_speed, current_speed);
    }

#ifdef DEBUG
                fprintf(stderr, "\n");
#endif
    check_cpu = check_therm = check_ac = false;
}

// handles the USR1 signal (stay at maximum performance)
void
usr1_handler(int)
{
    mode = SPEED_MAX;
    raise(SIGALRM);
}

// handles the USR2 signal (stay at minimum performance)
void
usr2_handler(int)
{
    mode = SPEED_MIN;
    raise(SIGALRM);
}

// handles the HUP signal (dynamically scale performance)
void
hup_handler(int)
{
    reset_times();
    mode = SPEED_DYNAMIC;
    check_cpu = true;
    raise(SIGALRM);
}

unsigned num_cpus = 0; // how many CPUs are we managing? 0 = autodetect

// restore  initial speed on program exit
unsigned saved_speed = 0;
char saved_governor[32];
pid_t * saved_pids;

void
term_handler(int which)
{
    if (saved_speed)
    {
        set_speed(saved_speed);
        write_line(GOVERNOR_FILE, "%s\n", saved_governor);
    }

    if (cpu == 0)
    {
        for (unsigned i = 1; i < num_cpus; i++)
            kill(saved_pids[i], which);
    }

    raise(which);
}

long
get_int(const char * s)
{
    char * end;
    long r = strtol(s, &end, 10);
    if (*end != '\0')
        die(false, "Not an integer: %s", s);
    if (errno == ERANGE)
        die(false, "Number is out of range: %s", s);

    return r;
}

char *
dup_cpu_str(const char *s)
{
    char buf[256];
    snprintf(buf, sizeof buf, s, cpu);
    return strdup(buf);
}

int
main(int argc, char * argv[])
{
    const char * const NAME = basename(strdup(argv[0]));

    // intervals (in tenths of a second) at which we poll
    unsigned interval = 20; // 2 seconds
    unsigned therm_interval = 10; // 1 second
    unsigned ac_interval = 50; // 5 seconds

    bool daemonize = false; // should the process daemonize itself?
    bool save_state = false; // save current speed and restore on exit
#ifdef DEBUG
    bool fake_cpu = false; // use fake CPUs for testing
#endif

    // parse argv
    for(int i = 1; i < argc; i++)
    {
        if(!strcmp(argv[i], "-d"))
            daemonize = true;
        else if(!strcmp(argv[i], "-r"))
            save_state = true;
        else if(!strcmp(argv[i], "-C"))
            max_speed_on_ac = true;
        else if(!strcmp(argv[i], "-D"))
            min_speed_on_battery = false;
        else if(!strcmp(argv[i], "-i"))
        {
            if (argc <= i + 1)
                die(false, "The \'-i\' option must be followed by an interval in tenths of a second");
            interval = get_int(argv[++i]);
#ifdef DEBUG
            fprintf(stderr, "CPU interval is %u\n", interval);
#endif
        }
	else if(!strcmp(argv[i], "-n"))
	    include_nice = true;
        else if(!strcmp(argv[i], "-p"))
        {
            if (argc <= i + 2)
                die(false, "The \'-p\' option must be followed by 2 integers");
            clock_up_idle_fast = get_int(argv[++i]);
            idle_threshold = get_int(argv[++i]);
#ifdef DEBUG
            fprintf(stderr, "Triggers are %u %u\n",clock_up_idle_fast, idle_threshold);
#endif
        }
        else if (!strcmp(argv[i], "-t"))
        {
            if (argc <= i + 2)
                die(false, "The \'-t\' option must be followed by a filename and a temperature");
            temperature_filename = argv[++i];
            max_temperature = get_int(argv[++i]);
        }
        else if (!strcmp(argv[i], "-a"))
        {
            if (argc <= i + 1)
                die(false, "The \'-a\' option must be followed by a filename");
            ac_filename = argv[++i];
        }
        else if(!strcmp(argv[i], "-m"))
        {
            if (argc <= i + 1)
                die(false, "The \'-m\' option must be followed by a minimum speed in KHz");
            min_speed = get_int(argv[++i]);
#ifdef DEBUG
            fprintf(stderr, "Minimum speed allowed by user: %uKHz\n", min_speed);
#endif
        }
        else if(!strcmp(argv[i], "-M"))
        {
            if (argc <= i + 1)
                die(false, "The \'-M\' option must be followed by a maximum speed in KHz");
            max_speed = get_int(argv[++i]);
#ifdef DEBUG
            fprintf(stderr, "Maximum speed allowed by user: %uKHz\n", max_speed);
#endif
        }
        else if(!strcmp(argv[i], "-T"))
        {
            if (argc <= i + 1)
                die(false, "The \'-T\' option must be followed by an interval in tenths of a second");
            therm_interval = get_int(argv[++i]);
#ifdef DEBUG
            fprintf(stderr, "Thermal interval is %u\n", therm_interval);
#endif
        }
        else if(!strcmp(argv[i], "-A"))
        {
            if (argc <= i + 1)
                die(false, "The \'-A\' option must be followed by an interval in tenths of a second");
            ac_interval = get_int(argv[++i]);
#ifdef DEBUG
            fprintf(stderr, "AC interval is %u\n", ac_interval);
#endif
        }
        else if(!strcmp(argv[i], "-s"))
        {
            if (argc <= i + 1)
                die(false, "The \'-s\' option must be followed by a CPU number");
            cpu = get_int(argv[++i]);
            num_cpus = 1;
#ifdef DEBUG
            fprintf(stderr, "Managing only CPU #%u\n", cpu);
#endif
        }
#ifdef DEBUG
        else if(!strcmp(argv[i], "-f"))
        {
            if (argc <= i + 1)
                die(false, "The \'-f\' option must be followed by the number of fake CPUs");
            if (fake_cpu)
                die(false, "The \'-f\' option may only be used once");
            if (num_cpus)
                die(false, "The \'-f\' option must come before the \'-s\' option");
            num_cpus = get_int(argv[++i]);
            fake_cpu = true;
            fprintf(stderr, "%u fake CPU(s) detected\n", num_cpus);
        }
#endif
        else
        {
            fprintf(
                stderr,
                "%s v%s\n"
                "\n"
                "This program monitors the system's idle percentage and reduces or raises the\n"
                "CPUs\' clock speeds and voltages accordingly to minimize power consumption\n"
                "when idle and maximize performance when needed.  This is the default.\n"
                "\n"
                "The program may also optionally be configured to reduce the CPUs\' clock\n"
                "speeds if the temperature gets too high, NOT minimize their speeds if the\n"
                "computer's AC adapter is disconnected or maximize their speeds when the AC\n"
                "adapter is connected.\n"
                "\n"
                "By default this program will manage every CPU found in the system.\n"
                "\n"
                "Usage: %s [Options]\n"
                "\n"
                "    Options:\n"
                "        -d\n"
                "            Tells the process to daemonize itself (run in background).\n"
                "\n"
                "        -i <interval>\n"
                "            Sets the interval between idle percentage tests and possible speed\n"
                "            changes in tenths of a second (default is 20).\n"
                "\n"
		"        -n\n"
                "            Includes nice time as cpu time (off by default)\n"
                "\n"
                "        -p <fast up> <threshold>\n"
                "            Sets the CPU idle percentage thresholds.  <fast up> is the idle\n"
                "            percentage below which a CPU will be set to the highest possible\n"
                "            speed.  <threshold> is the idle percentage above which a CPU's\n"
                "            speed will be decreased and below which a CPU's speed will be\n"
                "            increased (defaults are 10 and 25).\n"
                "\n"
                "        -m <minimum speed>\n"
                "            Sets the minimum speed in KHz below which a CPU will not be set.\n"
                "\n"
                "        -M <maximum speed>\n"
                "            Sets the maximum speed in KHz above which a CPU will not be set.\n"
                "\n"
                "        -t <temp file> <maxtemp>\n"
                "            Sets the ACPI temperature file and the temperature at which CPUs\n"
                "            will be set to minimum speed.\n"
                "\n"
                "        -T <interval>\n"
                "            Sets the interval at which the temperature will be polled in\n"
                "            tenths  of a second (default is 10).\n"
                "            (Requires the \'-t\' option above.)\n"
                "\n"
                "        -a <AC file>\n"
                "            Sets the ACPI AC adapter state file and tells the program to set\n"
                "            the CPUs to minimum speed when the AC adapter is disconnected.\n"
                "            (This is the default but is changeable by the \'-D\' option below).\n"
                "\n"
                "        -A <interval>\n"
                "            Sets the interval at which the AC adapter state will be polled in\n"
                "            tenths  of a second (default is 50).\n"
                "            (Requires the \'-a\' option above.)\n"
                "\n"
                "        -C\n"
                "            Run at maximum speed when AC adapter is connected.\n"
                "            (Requires the \'-a\' option above.)\n"
                "\n"
                "        -D\n"
                "            Do NOT force minimum speed when AC adapter is disconnected.\n"
                "            (Requires the \'-a\' option above.)\n"
                "\n"
                "        -r\n"
                "            Restores previous speed on program exit.\n"
                "\n"
                "        -s <CPU>\n"
                "            Manage only a single CPU.  <CPU> specifies the number of the CPU\n"
                "            to manage and is in the range 0 to (n-1) where \'n\' is the number\n"
                "            of CPUs in the system.  Without this option the program creates\n"
                "            copies of itself to manage every CPU in the system.\n"
                "\n"
                "    To have a CPU stay at the highest clock speed to maximize performance send\n"
                "    the process controlling that CPU the SIGUSR1 signal.\n"
                 "\n"
                "    To have a CPU stay at the lowest clock speed to maximize battery life send\n"
                "    the process controlling that CPU the SIGUSR2 signal.\n"
                 "\n"
                "    To resume having a CPU\'s clock speed dynamically scaled send the process\n"
                "    controlling that CPU the SIGHUP signal.\n"
                 "\n"
                "Author:\n"
                "    Carl Thompson - cet [at] carlthompson.net  (copyright 2002 - 2005)\n"
                "\n"
                , NAME, VERSION, NAME
            );
            exit(0);
        }
    }

    // set up signal handling to ignore SIGCHLD signals
    // uncommenting this could cause TERM signals to be sent to the wrong
    // processes on exit so we accept possible <defunct> processes
//     struct sigaction child_action;
//     sigemptyset(&child_action.sa_mask);
//     child_action.sa_flags = 0;
//     child_action.sa_handler = SIG_IGN;
//     sigaction(SIGCHLD, &child_action, 0);

    // get number of CPUs
    if (!num_cpus)
    {
        num_cpus = get_nprocs_conf();
#ifdef DEBUG
        fprintf(stderr, "%u CPU(s) detected.\n", num_cpus);
#endif
    }

    saved_pids = (pid_t *)calloc(num_cpus, sizeof(pid_t));

    // run in background if requested
    if (daemonize)
        daemon(0, 0);

    // fork() a process to handle each CPU
    for (unsigned i = 1; i < num_cpus; i++)
    {
        if ( !(saved_pids[i] = fork()) )
        {
            cpu = i;
            break;
        }
    }

    // if possible, set the proccess title to a more descriptive version
    if (num_cpus > 1)
    {
        char tmp[32];
        if ((size_t)snprintf(tmp, sizeof tmp, "%s#%u", NAME, cpu) <= strlen(argv[0]))
            strcpy(argv[0], tmp);
    }

    STAT_FILE = PROC_STAT_FILE;
#ifdef WITH_PROC
    struct utsname un;
    uname(&un);
    if (!strncmp(un.release, "2.4", 3))
    {
        MIN_SPEED_FILE = dup_cpu_str(PROC_MIN_SPEED_FILE);
        MAX_SPEED_FILE = dup_cpu_str(PROC_MAX_SPEED_FILE);
        CURRENT_SPEED_FILE = dup_cpu_str(PROC_CURRENT_SPEED_FILE);
        GOVERNOR_FILE = PROC_GOVERNOR_FILE;
        USERSPACE = PROC_USERSPACE;
    }
    else
#endif
    {
        MIN_SPEED_FILE = dup_cpu_str(SYSFS_MIN_SPEED_FILE);
        MAX_SPEED_FILE = dup_cpu_str(SYSFS_MAX_SPEED_FILE);
        CURRENT_SPEED_FILE = dup_cpu_str(SYSFS_CURRENT_SPEED_FILE);
        GOVERNOR_FILE = dup_cpu_str(SYSFS_GOVERNOR_FILE);
        USERSPACE = SYSFS_USERSPACE;
    }
#ifdef DEBUG
    if (fake_cpu)
    {
         STAT_FILE = FAKE_STAT_FILE;
        MIN_SPEED_FILE = dup_cpu_str(FAKE_MIN_SPEED_FILE);
        MAX_SPEED_FILE = dup_cpu_str(FAKE_MAX_SPEED_FILE);
        CURRENT_SPEED_FILE = dup_cpu_str(FAKE_CURRENT_SPEED_FILE);
        GOVERNOR_FILE = dup_cpu_str(FAKE_GOVERNOR_FILE);
        USERSPACE = FAKE_USERSPACE;
    }
#endif

    // save current speed if necessary
    if (save_state)
    {
        saved_speed = get_speed();
        read_line(GOVERNOR_FILE, saved_governor, sizeof saved_governor);
    }

   // use the userspace governor
    write_line(GOVERNOR_FILE, "%s\n", USERSPACE);

    if (access(CURRENT_SPEED_FILE, W_OK) < 0)
        die(true, "Cannot write to speed control file: %s", CURRENT_SPEED_FILE);

    // set up signal handling
    struct sigaction signal_action;

    // block all of our signals during each signal handler to avoid possible
    // race condition (not that it would really matter)
    sigemptyset(&signal_action.sa_mask);
    sigaddset(&signal_action.sa_mask, SIGALRM);
    sigaddset(&signal_action.sa_mask, SIGUSR1);
    sigaddset(&signal_action.sa_mask, SIGUSR2);
    sigaddset(&signal_action.sa_mask, SIGHUP);
    sigaddset(&signal_action.sa_mask, SIGTERM);
    sigaddset(&signal_action.sa_mask, SIGQUIT);
    sigaddset(&signal_action.sa_mask, SIGINT);
    signal_action.sa_flags = 0;

    // set the SIGALRM handler to our function
    signal_action.sa_handler = alarm_handler;
    sigaction(SIGALRM, &signal_action, 0);

    // set the SIGUSR1 handler to our function
    signal_action.sa_handler = usr1_handler;
    sigaction(SIGUSR1, &signal_action, 0);

    // set the SIGUSR2 handler to our function
    signal_action.sa_handler = usr2_handler;
    sigaction(SIGUSR2, &signal_action, 0);

    // set the HUP handler to our function
    signal_action.sa_handler = hup_handler;
    sigaction(SIGHUP, &signal_action, 0);

    // set up signal handling for terminate signal
    struct sigaction term_action;

    // block all of our signals during each signal handler to avoid possible
    // race condition (not that it would really matter)
    sigemptyset(&term_action.sa_mask);
    sigaddset(&term_action.sa_mask, SIGALRM);
    sigaddset(&term_action.sa_mask, SIGUSR1);
    sigaddset(&term_action.sa_mask, SIGUSR2);
    sigaddset(&term_action.sa_mask, SIGHUP);
    sigaddset(&term_action.sa_mask, SIGTERM);
    sigaddset(&term_action.sa_mask, SIGQUIT);
    sigaddset(&term_action.sa_mask, SIGINT);
    term_action.sa_flags = SA_ONESHOT;

    // set the TERM handler to our function
    term_action.sa_handler = term_handler;
    sigaction(SIGTERM, &term_action, 0);
    sigaction(SIGQUIT, &term_action, 0);
    sigaction(SIGINT, &term_action, 0);

    // reset the speed steps
    get_supported_speeds();
    // reset the time counters
    reset_times();

    // we dynamically scale speed to start
    mode = SPEED_DYNAMIC;

    unsigned counter = 0;
    unsigned cpu_interval_timeout = 0;
    unsigned therm_interval_timeout = 0;
    unsigned ac_interval_timeout = 0;
    unsigned next_timeout, d;
    timespec timeout;
    // main loop
    while (1)
    {
        if (counter == cpu_interval_timeout)
        {
            cpu_interval_timeout += interval;
            check_cpu = true;
        }
        next_timeout = cpu_interval_timeout;

        if (temperature_filename)
        {
            if (counter == therm_interval_timeout)
            {
                therm_interval_timeout += therm_interval;
                check_therm = true;
            }
            next_timeout = MIN(next_timeout, therm_interval_timeout);
        }

        if (ac_filename)
        {
            if (counter == ac_interval_timeout)
            {
                ac_interval_timeout += ac_interval;
                check_ac = true;
            }
            next_timeout = MIN(next_timeout, ac_interval_timeout);
        }

        if (check_cpu || check_ac || check_therm)
            raise(SIGALRM);

        d = next_timeout - counter;
        timeout.tv_sec = d / 10;
        timeout.tv_nsec = (d % 10) * 100000000LU;
        while (nanosleep(&timeout, &timeout) == -1 && errno == EINTR)
            ;

        counter = next_timeout;
    }

    // never reached
    return 0;
}

/* $Id: $ */
